@@ -1,6 +1,7 @@
-use std::fmt;
+use std::fmt::{self, Write};
 
 use owo_colors::{OwoColorize, Style};
+use unicode_width::UnicodeWidthStr;
 
 use crate::chain::Chain;
 use crate::handlers::theme::*;
@@ -21,6 +22,7 @@ See [crate::set_hook] for more details on customizing your global printer.
 #[derive(Debug, Clone)]
 pub struct GraphicalReportHandler {
     pub(crate) linkify_code: bool,
+    pub(crate) termwidth: usize,
     pub(crate) theme: GraphicalTheme,
 }
 
@@ -30,6 +32,7 @@ impl GraphicalReportHandler {
     pub fn new() -> Self {
         Self {
             linkify_code: true,
+            termwidth: 200,
             theme: GraphicalTheme::default(),
         }
     }
@@ -38,6 +41,7 @@ impl GraphicalReportHandler {
     pub fn new_themed(theme: GraphicalTheme) -> Self {
         Self {
             linkify_code: true,
+            termwidth: 200,
             theme,
         }
     }
@@ -47,6 +51,18 @@ impl GraphicalReportHandler {
         self.linkify_code = false;
         self
     }
+
+    /// Set a theme for this handler.
+    pub fn with_theme(mut self, theme: GraphicalTheme) -> Self {
+        self.theme = theme;
+        self
+    }
+
+    /// Sets the width to wrap the report at.
+    pub fn with_width(mut self, width: usize) -> Self {
+        self.termwidth = width;
+        self
+    }
 }
 
 impl Default for GraphicalReportHandler {
@@ -66,6 +82,7 @@ impl GraphicalReportHandler {
         diagnostic: &(dyn Diagnostic),
     ) -> fmt::Result {
         self.render_header(f, diagnostic)?;
+        writeln!(f)?;
         self.render_causes(f, diagnostic)?;
 
         if let Some(snippets) = diagnostic.snippets() {
@@ -80,12 +97,14 @@ impl GraphicalReportHandler {
     }
 
     fn render_header(&self, f: &mut impl fmt::Write, diagnostic: &(dyn Diagnostic)) -> fmt::Result {
-        let (severity_style, severity_icon) = match diagnostic.severity() {
-            Some(Severity::Error) | None => (self.theme.styles.error, self.theme.characters.x),
-            Some(Severity::Warning) => (self.theme.styles.warning, self.theme.characters.warning),
-            Some(Severity::Advice) => (self.theme.styles.advice, self.theme.characters.point_right),
-        };
-        write!(f, "{}", self.theme.characters.hbar.to_string().repeat(4))?;
+        let mut header = String::new();
+        let mut width = 0;
+        write!(
+            header,
+            "{}",
+            self.theme.characters.hbar.to_string().repeat(4)
+        )?;
+        width += UnicodeWidthStr::width(&header[..]);
         if self.linkify_code && diagnostic.url().is_some() && diagnostic.code().is_some() {
             let url = diagnostic.url().unwrap(); // safe
             let code = format!(
@@ -95,35 +114,43 @@ impl GraphicalReportHandler {
                     .expect("MIETTE BUG: already got checked for None")
             );
             let link = format!("\u{1b}]8;;{}\u{1b}\\{}\u{1b}]8;;\u{1b}\\", url, code);
-            write!(f, "[{}]", link.style(self.theme.styles.code))?;
+            write!(header, "[{}]", link.style(self.theme.styles.code))?;
+            width += UnicodeWidthStr::width(&url.to_string()[..])
+                + UnicodeWidthStr::width(&code[..])
+                + 2;
         } else if let Some(code) = diagnostic.code() {
-            write!(f, "[{}]", code.style(self.theme.styles.code))?;
-        }
-        writeln!(f, "{}", self.theme.characters.hbar.to_string().repeat(20),)?;
-        writeln!(f)?;
-        write!(f, "    {} ", severity_icon.style(severity_style))?;
-        if diagnostic.source().is_some() {
-            write!(f, "{}", self.theme.characters.mtop.style(severity_style))?;
-        } else {
-            write!(f, "{}", self.theme.characters.hbar.style(severity_style))?;
+            write!(header, "[{}]", code.style(self.theme.styles.code))?;
+            width += UnicodeWidthStr::width(&code.to_string()[..]) + 2;
         }
         writeln!(
             f,
-            "{}{} {}",
-            self.theme.characters.hbar.style(severity_style),
-            self.theme.characters.rarrow.style(severity_style),
-            diagnostic,
+            "{}{}",
+            header,
+            self.theme.characters.hbar.to_string().repeat(
+                self.termwidth
+                    .saturating_sub(width)
+                    .saturating_sub("Error: ".len())
+            ),
         )?;
         Ok(())
     }
 
     fn render_causes(&self, f: &mut impl fmt::Write, diagnostic: &(dyn Diagnostic)) -> fmt::Result {
-        let severity_style = match diagnostic.severity() {
-            Some(Severity::Error) | None => self.theme.styles.error,
-            Some(Severity::Warning) => self.theme.styles.warning,
-            Some(Severity::Advice) => self.theme.styles.advice,
+        let (severity_style, severity_icon) = match diagnostic.severity() {
+            Some(Severity::Error) | None => (self.theme.styles.error, self.theme.characters.x),
+            Some(Severity::Warning) => (self.theme.styles.warning, self.theme.characters.warning),
+            Some(Severity::Advice) => (self.theme.styles.advice, self.theme.characters.point_right),
         };
 
+        let initial_indent = format!("    {} ", severity_icon.style(severity_style));
+        let rest_indent = format!("    {} ", self.theme.characters.vbar.style(severity_style));
+        let width = self.termwidth.saturating_sub(4);
+        let opts = textwrap::Options::new(width)
+            .initial_indent(&initial_indent)
+            .subsequent_indent(&rest_indent);
+
+        writeln!(f, "{}", textwrap::fill(&diagnostic.to_string(), opts))?;
+
         if let Some(cause) = diagnostic.source() {
             let mut cause_iter = Chain::new(cause).peekable();
             while let Some(error) = cause_iter.next() {
@@ -132,13 +159,19 @@ impl GraphicalReportHandler {
                 } else {
                     self.theme.characters.lbot
                 };
-                let prefix = format!(
-                    "      {}{}{}",
+                let initial_indent = format!(
+                    "    {}{}{} ",
                     char, self.theme.characters.hbar, self.theme.characters.rarrow
                 )
                 .style(severity_style)
                 .to_string();
-                writeln!(f, "{} {}", prefix, error)?;
+                let rest_indent = format!("    {}   ", self.theme.characters.vbar)
+                    .style(severity_style)
+                    .to_string();
+                let opts = textwrap::Options::new(width)
+                    .initial_indent(&initial_indent)
+                    .subsequent_indent(&rest_indent);
+                writeln!(f, "{}", textwrap::fill(&error.to_string(), opts))?;
             }
         }
 
@@ -149,7 +182,12 @@ impl GraphicalReportHandler {
         if let Some(help) = diagnostic.help() {
             let help = help.style(self.theme.styles.help);
             writeln!(f)?;
-            writeln!(f, "    {} {}", self.theme.characters.fyi, help)?;
+            let width = self.termwidth.saturating_sub(4);
+            let initial_indent = format!("    {} ", self.theme.characters.fyi);
+            let opts = textwrap::Options::new(width)
+                .initial_indent(&initial_indent)
+                .subsequent_indent("      ");
+            writeln!(f, "{}", textwrap::fill(&help.to_string(), opts))?;
         }
         Ok(())
     }
@@ -554,7 +592,7 @@ impl Line {
 
     fn span_applies(&self, span: &FancySpan) -> bool {
         // Span starts in this line
-        (span.offset() >= self.offset && span.offset() <= self.offset +self.length)
+        (span.offset() >= self.offset && span.offset() < self.offset +self.length)
         // Span passes through this line
         || (span.offset() < self.offset && span.offset() + span.len() > self.offset + self.length) //todo
         // Span ends on this line
